use crate::post::{Post, Posts};
use anyhow::Result;
use base64::{engine::GeneralPurpose, Engine};
use reqwest::header::{HeaderMap, HeaderValue, AUTHORIZATION, USER_AGENT};
use reqwest::Response;

#[derive(Clone, Debug)]
pub struct Client<'a> {
    auth: Authentication<'a>,
    useragent: &'a str,
    host: &'static str,
    http_client: reqwest::Client,
}

impl<'a> Client<'a> {
    pub fn new(auth: Authentication<'a>, useragent: &'a str) -> Result<Self> {
        let mut header_map = HeaderMap::new();
        header_map.append(USER_AGENT, HeaderValue::from_str(useragent)?);
        if let Authentication::Authorized { username, apikey } = auth {
            let authorization = Self::get_authorization_value(username, apikey);
            header_map.append(AUTHORIZATION, HeaderValue::from_str(&authorization)?);
        }

        let http_client = reqwest::Client::builder()
            .default_headers(header_map)
            .build()?;

        Ok(Client {
            auth,
            useragent,
            host: "https://e621.net",
            http_client,
        })
    }

    fn get_authorization_value(username: &'a str, apikey: &'a str) -> String {
        let base64_engine = GeneralPurpose::new(&base64::alphabet::STANDARD, Default::default());
        base64_engine.encode(format!("{username}:{apikey}"))
    }

    async fn list_posts_raw(
        &mut self,
        limit: Option<u16>,
        tags: Option<Vec<String>>,
        page: Option<u32>,
    ) -> Result<Response> {
        let mut url = url::Url::parse(format!("{}/posts.json", self.host).as_str())?;

        let mut query_params = Vec::new();
        if let Some(limit) = limit {
            query_params.push(format!("limit={limit}"));
        }

        if let Some(page) = page {
            query_params.push(format!("page={page}"));
        }

        if let Some(tags) = tags {
            if tags.len() > 0 {
                query_params.push(format!("tags={}", tags.join(",")));
            }
        }

        url.set_query(Some(&query_params.join("&")));

        Ok(self.http_client.get(url.as_str()).send().await?)
    }

    pub async fn list_posts(
        &mut self,
        limit: Option<u16>,
        tags: Option<Vec<String>>,
        page: Option<u32>,
    ) -> Result<Vec<Post>> {
        let res = self.list_posts_raw(limit, tags, page).await?;
        let text = res.text().await?;

        /*
        let mut debug_file = std::fs::OpenOptions::new()
            .write(true)
            .create(true)
            .open("./debug.json")?;
        debug_file.write_all(text.as_bytes())?;
        */

        Ok(serde_json::from_str::<Posts>(text.as_str())?.into())
    }
}

#[derive(Clone, Copy, Debug)]
pub enum Authentication<'a> {
    Authorized { username: &'a str, apikey: &'a str },
    Unauthorized,
}
